package com.namasoft.common.flatobjects;

import com.namasoft.common.constants.*;
import com.namasoft.common.exceptions.NaMaServiceExcepption;
import com.namasoft.common.flatobjects.http.*;
import com.namasoft.common.layout.edit.NaMaText;
import com.namasoft.common.urlutils.*;
import com.namasoft.common.utilities.*;
import com.namasoft.common.utils.*;
import com.namasoft.common.utils.translation.TranslationUtil;
import org.apache.commons.io.IOUtils;

import java.io.*;
import java.math.BigDecimal;
import java.net.*;
import java.nio.charset.StandardCharsets;
import java.text.MessageFormat;
import java.util.*;
import java.util.function.*;
import java.util.stream.Collectors;

import static com.namasoft.common.constants.PlaceTokens.*;
import static com.namasoft.common.utilities.NaMaKeyValuePairsTokenizer.pair;

public class ComplexRenderer
{
	public static final String START_ATTACH_DATA_TAG = "<attachData>";
	public static final String END_ATTACH_DATA_TAG = "</attachData>";
	public static final String END_MSG_SUBJECT = "</endMsgSubject>";
	public static final String START_MSG_SUBJECT = "<startMsgSubject>";
	public static final String END_SEND_TO = "</endSendTo>";
	public static final String START_SEND_TO = "<startSendTo>";
	public static final String END_MSG = "</endMsg>";
	public static final String START_NEW_MSG = "<startNewMsg>";
	private static final String OPEN_TABLE = "\n<table border='1' cellspacing='0'>";
	private String text;
	private RenderNode currentNode;
	private RootNode rootNode;
	private final Map<String, Integer> counters = new HashMap<>();
	private boolean directLink = false;
	private boolean shortLinks = false;
	private boolean dynamicCriteria = false;
	private String appURL;
	private String originalAppURL;
	public static IComplexRendererGlobalHelper HELPER = null;
	private boolean useCSSFriendlyBrackets = false;

	public ComplexRenderer(String text)
	{
		this.text = text;
	}

	public boolean isDirectLink()
	{
		return directLink;
	}

	public ComplexRenderer directLink(boolean directLink)
	{
		this.directLink = directLink;
		return this;
	}

	public ComplexRenderer dynamicCriteria(boolean dynamicCriteria)
	{
		setDynamicCriteria(dynamicCriteria);
		return this;
	}

	public void shortLinks(boolean shortLinks)
	{
		this.shortLinks = shortLinks;
	}

	public boolean isDynamicCriteria()
	{
		return dynamicCriteria;
	}

	public void setDynamicCriteria(boolean dynamicCriteria)
	{
		this.dynamicCriteria = dynamicCriteria;
	}

	public static StringBuilder render(String template, StringBuilder result, RendererPropertyResolver resolver)
	{
		if (result == null)
			result = new StringBuilder();
		return result.append(parse(template).render(resolver));
	}

	public boolean isShortLinks()
	{
		return shortLinks;
	}

	public List<String> collectVariables()
	{
		List<String> variables = new ArrayList<>();
		rootNode.visit(n -> variables.addAll(n.collectVariables()));
		return variables.stream().filter(ObjectChecker::isNotEmptyOrNull).filter(s -> !s.contains("\""))
				.filter(s -> ObjectChecker.tryParseDecimal(s) == null).collect(Collectors.toList());
	}

	public abstract class RenderNode
	{
		public ComplexRenderer renderer()
		{
			return ComplexRenderer.this;
		}

		private final List<RenderNode> nodes = new ArrayList<>();
		private String textPart = "";
		protected RenderNode parentNode;
		private Boolean insideCreator = null;

		public RenderNode getParentNode()
		{
			return parentNode;
		}

		public List<RenderNode> getNodes()
		{
			return nodes;
		}

		public List<RenderNode> getNodesForRendering()
		{
			return getNodes();
		}

		public String getTextPart()
		{
			return textPart;
		}

		public void setTextPart(String textPart)
		{
			this.textPart = textPart;
		}

		public void consume(char c)
		{
			textPart += c;
		}

		public void addSubNode(RenderNode subNode)
		{
			if (isLeafNode())
			{
				parentNode.addSubNode(subNode);
			}
			else
			{
				if (shouldSubNodeEndThis(subNode))
				{
					end("end");
					currentNode.addSubNode(subNode);
				}
				else
				{
					nodes.add(subNode);
					subNode.parentNode = this;
				}
			}
			currentNode = subNode;
		}

		public boolean shouldSubNodeEndThis(RenderNode subNode)
		{
			return false;
		}

		public abstract boolean isLeafNode();

		public void renderSubNodes(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			for (RenderNode node : getNodesForRendering())
			{
				node.render(builder, helper, currentLineNumber);
			}
		}

		public String renderSubNodes(RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder builder = new StringBuilder();
			for (RenderNode node : getNodes())
			{
				node.render(builder, helper, currentLineNumber);
			}
			return builder.toString();
		}

		public boolean canHaveText()
		{
			return false;
		}

		public void end(String nodeContent)
		{
			if (isLeafNode())
			{
				parentNode.end(nodeContent);
			}
			else
			{
				if (ObjectChecker.areEqual(nodeContent, "end") || isMyEnder(nodeContent))
					currentNode = parentNode;
				else
					parentNode.end(nodeContent);
			}
		}

		protected boolean isMyEnder(String nodeContent)
		{
			return false;
		}

		public void toString(StringBuilder builder, int indentationLevel)
		{
			builder.append("\n").append(indent(indentationLevel)).append(describe());
			for (RenderNode node : nodes)
			{
				node.toString(builder, indentationLevel + 1);
			}
			if (!nodes.isEmpty())
				builder.append("\n").append(indent(indentationLevel)).append("end:").append(describe());
		}

		private String indent(int indentationLevel)
		{
			StringBuilder str = new StringBuilder();
			for (int i = 0; i < indentationLevel; i++)
				str.append("\t");
			return str.toString();
		}

		protected abstract String describe();

		public abstract void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber);

		public void visit(Consumer<RenderNode> visitor)
		{
			visitor.accept(this);
			for (RenderNode node : getNodes())
			{
				node.visit(visitor);
			}
		}

		protected boolean insideCreator()
		{
			if (insideCreator != null)
				return insideCreator;
			insideCreator = false;
			RenderNode parent = this.parentNode;
			while (parent != null)
			{
				if (parent instanceof CreatorNode)
				{
					insideCreator = true;
					break;
				}
				else if (parent instanceof CreatorMultiValueNode)
				{
					break;
				}
				parent = parent.parentNode;
			}
			return insideCreator;
		}

		public boolean instanceOfOrChildOfType(Class<? extends RenderNode> nodeClass)
		{
			ComplexRenderer.RenderNode current = this;
			while (current != null)
			{
				if (nodeClass.isInstance(current))
					return true;
				current = current.getParentNode();
			}
			return false;
		}

		public List<String> collectVariables()
		{
			return Collections.emptyList();
		}
	}

	private class EnterNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{enter}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append("<br/>");
		}
	}

	private class OpenInNewWindowNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{openinnewwindow}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(CommonConstants.OPEN_LINK_IN_NEW_WINDOW);
		}
	}

	private class StartMsgNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{openmsg}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(START_NEW_MSG);
		}
	}

	private class EndMsgNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{closemsg}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(END_MSG);
		}
	}

	private class OpenTableNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{opentable}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(OPEN_TABLE);
		}
	}

	private class CloseTableNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{closetable}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append("\n</table>");
		}
	}

	private class DirectLinksNode extends RenderNode
	{
		public DirectLinksNode()
		{
			directLink(true);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{directlinks}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			directLink(true);
		}
	}

	private class AppUrlNode extends RenderNode
	{
		private final String nodeContent;
		protected final String url;

		public AppUrlNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.url = StringUtils.getInnerString(nodeContent, "(", ")");
			if (ObjectChecker.isEmptyOrNull(originalAppURL))
				originalAppURL = HELPER.getGuiServerURL();
			HELPER.setGuiServerURL(this.url);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (!url.startsWith("\"") && url.contains(":"))
				appURL = url;
			else
				appURL = evalConstantOrVar(url, helper, currentLineNumber, insideCreator(), this);
		}
	}

	private class ShortLinksNode extends RenderNode
	{
		public ShortLinksNode()
		{
			shortLinks(true);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{shortlinks}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			shortLinks(true);
		}
	}

	private class DynamicCriteriaNode extends RenderNode
	{
		public DynamicCriteriaNode()
		{
			setDynamicCriteria(true);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{dynamiccriteria}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			setDynamicCriteria(true);
		}
	}

	private class CommentNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{comment}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endcomment");
		}
	}

	private class RootNode extends RenderNode
	{
		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			renderSubNodes(builder, helper, currentLineNumber);
		}
	}

	private class TextNode extends RenderNode
	{
		@Override
		public boolean canHaveText()
		{
			return true;
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return '"' + getTextPart() + '"';
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (insideCreator())
				builder.append(getTextPart().trim());
			else
				builder.append(getTextPart());
		}
	}

	private class VarNode extends RenderNode
	{
		private final String varName;
		private Integer index;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(varName);
		}

		public VarNode(String nodeContent)
		{
			List<String> parts = StringUtils.csvLineToList(nodeContent);
			this.varName = parts.get(0);
			if (parts.size() > 1)
			{
				index = ObjectChecker.tryParseInt(parts.get(1));
				if (index == null)
					index = 0;
				else
					index = index - 1;
			}
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + varName + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (isDynamicCriteria())
				builder.append(helper.marshalInString(helper.resolveObject(varName, ObjectChecker.getFirstNotNullObj(index, currentLineNumber))));
			else
				builder.append(helper.resolve(varName, ObjectChecker.getFirstNotNullObj(index, currentLineNumber), insideCreator(), this));
		}
	}

	private List<HttpRequest> httpRequests;

	public List<HttpRequest> getHttpRequests()
	{
		if(httpRequests == null)
			httpRequests = new ArrayList<>();
		return httpRequests;
	}

	private HttpRequest currentRequest;
	private HttpRequestBodyPartHandler currentBodyPartHandler;

	private HttpRequestBodyPartHandler currentBodyPartHandler()
	{
		return currentBodyPartHandler == null ? currentRequest : currentBodyPartHandler;
	}
	private abstract class AbsHttpRequestNode extends RenderNode
	{
		protected final String nodeContent;

		public AbsHttpRequestNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, enderStr());
		}

		protected abstract String enderStr();
	}

	private class HttpRequestNode extends AbsHttpRequestNode
	{
		public HttpRequestNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (httpRequests == null)
				httpRequests = new ArrayList<>();
			httpRequests.add(currentRequest = new HttpRequest());
			renderSubNodes(builder, helper, currentLineNumber);
		}

		@Override
		protected String enderStr()
		{
			return "endrequest";
		}
	}

	private class HttpRequestMethodNode extends AbsHttpRequestNode
	{
		public HttpRequestMethodNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setMethod(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endmethod";
		}
	}

	private class HttpRequestUrlNode extends AbsHttpRequestNode
	{
		public HttpRequestUrlNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setUrl(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endurl";
		}
	}

	private class HttpRequestContentTypeNode extends AbsHttpRequestNode
	{
		public HttpRequestContentTypeNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setContentType(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endcontenttype";
		}
	}

	private class HttpRequestCharSetNode extends AbsHttpRequestNode
	{
		public HttpRequestCharSetNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setCharSet(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endcharset";
		}
	}

	private class HttpRequestRelatedToId1Node extends AbsHttpRequestNode
	{
		public HttpRequestRelatedToId1Node(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setRelatedToId1(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endrelatedtoid1";
		}
	}

	private class HttpRequestRelatedToId2Node extends AbsHttpRequestNode
	{
		public HttpRequestRelatedToId2Node(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setRelatedToId2(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endrelatedtoid2";
		}
	}

	private class HttpRequestDescription1Node extends AbsHttpRequestNode
	{
		public HttpRequestDescription1Node(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setDescription1(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "enddescription1";
		}
	}

	private class HttpRequestDescription2Node extends AbsHttpRequestNode
	{
		public HttpRequestDescription2Node(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setDescription2(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "enddescription2";
		}
	}

	private class HttpRequestHeaderNameNode extends AbsHttpRequestNode
	{

		public HttpRequestHeaderNameNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.addHeader(renderSubNodes(helper, currentLineNumber), null);
		}

		@Override
		protected String enderStr()
		{
			return "endheadername";
		}
	}

	private class HttpRequestHeaderValueNode extends AbsHttpRequestNode
	{

		public HttpRequestHeaderValueNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			CollectionsUtility.getLast(currentRequest.getHeaders()).setValue(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endheadervalue";
		}
	}

	private class HttpRequestParameterNameNode extends AbsHttpRequestNode
	{

		public HttpRequestParameterNameNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.addParameter(renderSubNodes(helper, currentLineNumber), null);
		}

		@Override
		protected String enderStr()
		{
			return "endparamname";
		}
	}

	private class HttpRequestParameterValueNode extends AbsHttpRequestNode
	{

		public HttpRequestParameterValueNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			CollectionsUtility.getLast(currentRequest.getParameters()).setValue(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endparamvalue";
		}
	}

	private class HttpRequestBodyPartNameNode extends AbsHttpRequestNode
	{
		public HttpRequestBodyPartNameNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentBodyPartHandler().addBodyPart(renderSubNodes(helper, currentLineNumber), null);
		}

		@Override
		protected String enderStr()
		{
			return "endbodypartname";
		}
	}

	//todo consider creating body part array class later, it should have multiple bodypartvalue nodes, and each one adds an object to the array
	private class HttpRequestBodyPartNode extends AbsHttpRequestNode
	{
		private final String expression;

		public HttpRequestBodyPartNode(String nodeContent)
		{
			super(nodeContent);
			this.expression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			TreeMap<Object, Object> value = new TreeMap<>();
			currentBodyPartHandler().addBodyPart(evalConstantOrVar(expression, helper, currentLineNumber, false, this), value);
			HttpRequestBodyPartHandler lastHandler = currentBodyPartHandler;
			currentBodyPartHandler = new HttpRequestBodyPartHandler()
			{
				private String lastKey;

				@Override
				public void addBodyPart(String name, Object val)
				{
					value.put(lastKey = name, val);
				}

				@Override
				public void setLastBodyPartValue(String val)
				{
					value.put(lastKey, val);
				}
			};
			renderSubNodes(builder, helper, currentLineNumber);
			currentBodyPartHandler = lastHandler;
		}

		@Override
		protected String enderStr()
		{
			return "endbodypart";
		}
	}

	private class HttpRequestBodyPartValueNode extends AbsHttpRequestNode
	{
		public HttpRequestBodyPartValueNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentBodyPartHandler().setLastBodyPartValue(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endbodypartvalue";
		}
	}

	private class HttpRequestBodyNode extends AbsHttpRequestNode
	{
		public HttpRequestBodyNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			currentRequest.setBody(renderSubNodes(helper, currentLineNumber));
		}

		@Override
		protected String enderStr()
		{
			return "endbody";
		}
	}

	private class LoopNode extends RenderNode
	{

		private final String varNameAndIndices;
		private String loopDetailName;
		private Integer startIndex;
		private Integer endIndex;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(loopDetailName);
		}

		public LoopNode(String nodeContent)
		{
			this.varNameAndIndices = nodeContent;
			String params = StringUtils.getInnerString(varNameAndIndices, "(", ")");
			if (ObjectChecker.isEmptyOrNull(params))
				return;
			List<String> parts = StringUtils.csvLineToList(params);
			if (ObjectChecker.isEmptyOrNull(parts))
				return;
			if (ObjectChecker.areEqual(parts.get(0), "last"))
				startIndex = Integer.MAX_VALUE;
			else if (ObjectChecker.tryParseInt(parts.get(0)) != null)
				startIndex = ObjectChecker.tryParseInt(parts.get(0));
			else
				loopDetailName = parts.get(0);

			if (parts.size() <= 1)
				return;
			if (ObjectChecker.isNotEmptyOrNull(startIndex))
			{
				endIndex = ObjectChecker.tryParseInt(parts.get(1));
			}
			else
			{
				if (ObjectChecker.areEqual(parts.get(1), "last"))
					startIndex = Integer.MAX_VALUE;
				else
					startIndex = ObjectChecker.tryParseInt(parts.get(1));
			}
			if (parts.size() <= 2)
				return;
			endIndex = ObjectChecker.tryParseInt(parts.get(2));
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + varNameAndIndices + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			int listSize = helper.calcSizeForLoop(loopDetailName);
			int startIndex = 0;
			if (this.startIndex != null)
			{
				if (ObjectChecker.areEqual(this.startIndex, Integer.MAX_VALUE))
					startIndex = listSize - 1;
				else
					startIndex = this.startIndex - 1;
			}
			int endIndex = listSize - 1;
			if (this.endIndex != null)
				endIndex = this.endIndex - 1;
			helper.loopStart(builder);
			for (int i = startIndex; i <= endIndex && i < listSize; i++)
			{
				helper.loopIterationStart(builder);
				renderSubNodes(builder, helper, i);
				helper.loopIterationEnd(builder);
			}
			helper.loopEnd(builder);
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endloop");
		}
	}

	private class TempoNode extends RenderNode
	{

		private final String nodeContent;

		public TempoNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder tempo = new StringBuilder();
			renderSubNodes(tempo, helper, currentLineNumber);
			StringBuilder renderedTempo = ComplexRenderer.parse(tempo.toString()).render(helper);
			builder.append(renderedTempo);
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endtempo");
		}
	}

	private class IfEqualNode extends AbstractIfNode
	{
		protected final String leftSideVarName;
		protected final String rightSideVarName;
		protected final boolean numeric;

		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(leftSideVarName, rightSideVarName);
		}

		public IfEqualNode(String nodeContent)
		{
			super(nodeContent);
			this.numeric = nodeContent.toLowerCase().contains("ifnumber");
			String conditionBody = StringUtils.getInnerString(condition, "(", ")");
			this.leftSideVarName = StringUtils.substringBefore(conditionBody, ",");
			this.rightSideVarName = StringUtils.substringAfter(conditionBody, ",");
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		protected boolean checkCondition(Object leftSideValue, Object rightSideValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (numeric)
				return ObjectChecker.areEqual(helper.toDecimal(leftSideValue), helper.toDecimal(rightSideValue));
			return ObjectChecker.areEqual(ObjectChecker.toStringOrEmpty(leftSideValue), ObjectChecker.toStringOrEmpty(rightSideValue));
		}

		@Override
		protected boolean checkCondition(RendererPropertyResolver helper, int currentLineNumber)
		{
			Object leftSideValue = evalConstantOrVar(leftSideVarName, helper, currentLineNumber, insideCreator(), this);
			Object rightSideValue = evalConstantOrVar(rightSideVarName, helper, currentLineNumber, insideCreator(), this);
			return checkCondition(leftSideValue, rightSideValue, helper, currentLineNumber);
		}
	}

	private class PadRightNode extends PadLeftNode
	{
		public PadRightNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean appendSpacesToStart()
		{
			return false;
		}

		@Override
		protected int calcEndIndex(StringBuilder subBuilder)
		{
			return length;
		}

		@Override
		protected int clacStartIndex(StringBuilder subBuilder)
		{
			return 0;
		}
	}

	private class PadLeftNode extends RenderNode
	{
		protected final int length;
		private final String nodeContent;

		public PadLeftNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.length = Integer.parseInt(StringUtils.getInnerString(nodeContent, "(", ")"));
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder subBuilder = new StringBuilder();
			renderSubNodes(subBuilder, helper, currentLineNumber);
			if (subBuilder.length() > length)
			{
				builder.append(subBuilder.substring(clacStartIndex(subBuilder), calcEndIndex(subBuilder)));
			}
			else if (subBuilder.length() < length)
			{
				StringBuilder spaces = new StringBuilder();
				for (int i = subBuilder.length(); i < length; i++)
				{
					spaces.append(" ");
				}
				if (appendSpacesToStart())
					builder.append(spaces).append(subBuilder);
				else
					builder.append(subBuilder).append(spaces);
			}
			else
			{
				builder.append(subBuilder);
			}
		}

		protected int clacStartIndex(StringBuilder subBuilder)
		{
			return subBuilder.length() - length - 1;
		}

		protected int calcEndIndex(StringBuilder subBuilder)
		{
			return subBuilder.length();
		}

		protected boolean appendSpacesToStart()
		{
			return true;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endpad");
		}
	}

	private abstract class AbsCounterNode extends RenderNode
	{
		private final String nodeContent;
		private final String name;

		public AbsCounterNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.name = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (isIncrement())
				counters.compute(name, (k, o) -> ObjectChecker.toZeroIfNull(o) + 1);
			else if (isDecrement())
				counters.compute(name, (k, o) -> ObjectChecker.toZeroIfNull(o) - 1);
			else
				builder.append(ObjectChecker.toZeroIfNull(counters.get(name)));
		}

		protected boolean isIncrement()
		{
			return false;
		}

		protected boolean isDecrement()
		{
			return false;
		}
	}

	private class IncrementCounterNode extends AbsCounterNode
	{

		public IncrementCounterNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean isIncrement()
		{
			return true;
		}
	}

	private class DecrementCounterNode extends AbsCounterNode
	{

		public DecrementCounterNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean isDecrement()
		{
			return true;
		}
	}

	private class CounterValueNode extends AbsCounterNode
	{

		public CounterValueNode(String nodeContent)
		{
			super(nodeContent);
		}
	}

	private class IfNotEqualNode extends IfEqualNode
	{
		public IfNotEqualNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean checkCondition(Object leftSideValue, Object rightSideValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			return !super.checkCondition(leftSideValue, rightSideValue, helper, currentLineNumber);
		}
	}

	private class IfLessNode extends IfEqualNode
	{
		public IfLessNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean checkCondition(Object leftSideValue, Object rightSideValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.areAllEmptyOrNull(leftSideValue, rightSideValue))
				return false;
			return helper.compareValues(leftSideValue, rightSideValue, numeric) < 0;
		}
	}

	private class IfLessOrEqNode extends IfEqualNode
	{
		public IfLessOrEqNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean checkCondition(Object leftSideValue, Object rightSideValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.areAllEmptyOrNull(leftSideValue, rightSideValue))
				return false;
			return helper.compareValues(leftSideValue, rightSideValue, numeric) <= 0;
		}
	}

	private class IfGreaterNode extends IfEqualNode
	{
		public IfGreaterNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean checkCondition(Object leftSideValue, Object rightSideValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.areAllEmptyOrNull(leftSideValue, rightSideValue))
				return false;
			return helper.compareValues(leftSideValue, rightSideValue, numeric) > 0;
		}
	}

	private class IfGreaterOrEqNode extends IfEqualNode
	{
		public IfGreaterOrEqNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean checkCondition(Object leftSideValue, Object rightSideValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.areAllEmptyOrNull(leftSideValue, rightSideValue))
				return false;
			return helper.compareValues(leftSideValue, rightSideValue, numeric) >= 0;
		}
	}

	private abstract class AbstractIfNode extends RenderNode
	{
		protected final boolean elseNode;
		protected final String condition;

		public boolean acceptsElseNodes()
		{
			return true;
		}

		public AbstractIfNode(String nodeContent)
		{
			this.condition = nodeContent;
			this.elseNode = nodeContent.startsWith("else");
		}

		@Override
		public List<RenderNode> getNodesForRendering()
		{
			return getNodes().stream().filter(n -> !(n instanceof AbstractIfNode ifNode && ifNode.isElse())).toList();
		}

		public List<AbstractIfNode> getElseNodes()
		{
			return getNodes().stream().map(n -> n instanceof AbstractIfNode ifNode ? ifNode : null).filter(Objects::nonNull)
					.filter(AbstractIfNode::isElse).toList();
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (checkCondition(helper, currentLineNumber))
			{
				renderSubNodes(builder, helper, currentLineNumber);
			}
			else
			{
				for (AbstractIfNode elseNode : getElseNodes())
				{
					if (elseNode.checkCondition(helper, currentLineNumber))
					{
						//directly render else node, do not call render because it will check the condition again, might be expensive
						elseNode.renderSubNodes(builder, helper, currentLineNumber);
						break;
					}
				}
			}
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, isElse() ? "endelse" : "endif");
		}

		public boolean isElse()
		{
			return elseNode;
		}

		@Override
		protected String describe()
		{
			return "{" + condition + "}";
		}

		protected abstract boolean checkCondition(RendererPropertyResolver helper, int currentLineNumber);
	}

	private class IfNode extends AbstractIfNode
	{
		protected final String varName;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(varName);
		}

		public IfNode(String nodeContent)
		{
			super(nodeContent);
			this.varName = StringUtils.getInnerString(condition, "(", ")");
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected boolean checkCondition(RendererPropertyResolver helper, int currentLineNumber)
		{
			Object conditionValue = helper.resolveObject(varName, currentLineNumber);
			return checkCondition(conditionValue, helper, currentLineNumber);
		}

		protected boolean checkCondition(Object conditionValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			return helper.isTrueIfCondition(conditionValue);
		}
	}

	private class IfNotNode extends IfNode
	{
		public IfNotNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean checkCondition(Object conditionValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			return !super.checkCondition(conditionValue, helper, currentLineNumber);
		}
	}

	private class ElseNode extends IfNode
	{
		public ElseNode()
		{
			super("else");
		}

		@Override
		protected boolean checkCondition(Object conditionValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			return true;
		}

		@Override
		public boolean isElse()
		{
			return true;
		}

		@Override
		public boolean acceptsElseNodes()
		{
			return false;
		}
	}

	private class SubstringNode extends RenderNode
	{
		private final String nodeContent;
		protected String firstParameter;
		protected String secondParameter;
		protected String thirdParameter;

		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(firstParameter, secondParameter, thirdParameter);
		}

		public SubstringNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			String innerString = StringUtils.getInnerString(nodeContent, "(", ")");
			List<String> parts = StringUtils.csvLineToList(innerString);
			if (!parts.isEmpty())
				firstParameter = parts.get(0);
			if (parts.size() > 1)
				secondParameter = parts.get(1);
			if (parts.size() > 2)
				thirdParameter = parts.get(2);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.isEmptyOrNull(firstParameter))
				return;
			String str = ObjectChecker.toStringOrEmpty(evalConstantOrVar(firstParameter, helper, currentLineNumber, insideCreator(), this));
			if (ObjectChecker.isEmptyOrNull(str))
				return;
			String startStr = "0";
			if (ObjectChecker.isNotEmptyOrNull(secondParameter))
				startStr = ObjectChecker.toStringOrEmpty(evalConstantOrVar(secondParameter, helper, currentLineNumber, insideCreator(), this));
			String lenStr = str.length() + "";
			if (ObjectChecker.isNotEmptyOrNull(thirdParameter))
				lenStr = ObjectChecker.toStringOrEmpty(evalConstantOrVar(thirdParameter, helper, currentLineNumber, insideCreator(), this));
			int start = parseInt(startStr);
			start = Math.max(start, 0);
			int len = parseInt(lenStr);
			if (len < 0)
				len = str.length();
			int endIndex = Math.min(str.length(), start + len);
			builder.append(str, start, endIndex);
		}
	}

	private class TafqeetNode extends RenderNode
	{
		private final String nodeContent;
		protected String firstParameter;
		protected String secondParameter;

		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(firstParameter, secondParameter);
		}

		public TafqeetNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			String innerString = StringUtils.getInnerString(nodeContent, "(", ")");
			List<String> parts = StringUtils.csvLineToList(innerString);
			if (!parts.isEmpty())
				firstParameter = parts.get(0);
			if (parts.size() > 1)
				secondParameter = parts.get(1);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.isEmptyOrNull(firstParameter))
				return;
			helper.tafqeet(evalConstantOrVar(firstParameter, helper, currentLineNumber, insideCreator(), this),
					evalConstantOrVar(secondParameter, helper, currentLineNumber, insideCreator(), this), builder, getLanguage());
		}

		protected Language getLanguage()
		{
			return TranslationUtil.getCurrentLanguage();
		}
	}

	private class TafqeetArNode extends TafqeetNode
	{

		public TafqeetArNode(String nodeContent)
		{
			super(nodeContent);
		}

		protected Language getLanguage()
		{
			return Language.Arabic;
		}
	}

	private class RoundNode extends TafqeetNode
	{

		public RoundNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.isEmptyOrNull(firstParameter))
				return;
			helper.round(evalConstantOrVarToObj(firstParameter, helper, currentLineNumber),
					evalConstantOrVarToObj(secondParameter, helper, currentLineNumber), builder);
		}

	}

	private class FormatNumberNode extends TafqeetNode
	{

		public FormatNumberNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.isEmptyOrNull(firstParameter))
				return;
			helper.formatNumber(evalConstantOrVarToObj(firstParameter, helper, currentLineNumber),
					evalConstantOrVar(secondParameter, helper, currentLineNumber, insideCreator(), this), builder);
		}

	}

	private class FormatDateNode extends TafqeetNode
	{

		public FormatDateNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (ObjectChecker.isEmptyOrNull(firstParameter))
				return;
			helper.formatDate(evalConstantOrVarToObj(firstParameter, helper, currentLineNumber),
					evalConstantOrVar(secondParameter, helper, currentLineNumber, insideCreator(), this), builder);
		}

	}

	private class TafqeetEnNode extends TafqeetNode
	{

		public TafqeetEnNode(String nodeContent)
		{
			super(nodeContent);
		}

		protected Language getLanguage()
		{
			return Language.English;
		}
	}

	private class LeftNode extends SubstringNode
	{
		public LeftNode(String nodeContent)
		{
			super(nodeContent);
			thirdParameter = secondParameter;
			secondParameter = "0";
		}
	}

	private class RightNode extends SubstringNode
	{
		public RightNode(String nodeContent)
		{
			super(nodeContent);
			thirdParameter = secondParameter;
			secondParameter = "-";
		}
	}

	private int parseInt(String s)
	{
		return parseDecimal(s).intValue();
	}

	private BigDecimal parseDecimal(String s)
	{
		try
		{
			return new BigDecimal(s);
		}
		catch (Exception e)
		{
			return BigDecimal.ZERO;
		}
	}

	private class GroupHeaderNode extends IfNode
	{
		public GroupHeaderNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean acceptsElseNodes()
		{
			return false;
		}

		@Override
		protected boolean checkCondition(Object conditionValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (currentLineNumber == 0 || ObjectChecker.areNotEqual(helper.resolveObject(varName, currentLineNumber - 1), conditionValue))
				return true;
			return false;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endheader");
		}
	}

	private class GroupFooterNode extends IfNode
	{
		public GroupFooterNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean acceptsElseNodes()
		{
			return false;
		}

		@Override
		protected boolean checkCondition(Object conditionValue, RendererPropertyResolver helper, int currentLineNumber)
		{
			int listSize = helper.listSize(varName);
			if (listSize == 0 || listSize == currentLineNumber + 1 || ObjectChecker.areNotEqual(helper.resolveObject(varName, currentLineNumber + 1),
					conditionValue))
				return true;
			return false;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endfooter");
		}
	}

	private class RowNumberNode extends RenderNode
	{

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{@rownumber}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(currentLineNumber + 1);
		}
	}

	private class CreatorNode extends RenderNode
	{
		private final String nodeContent;
		private String entityTypeExpression;
		private String titleExpression;
		private String menuCode;
		private String targetExpression;
		private String viewCode;

		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(entityTypeExpression, titleExpression, menuCode, targetExpression, viewCode);
		}

		private CreatorNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			List<String> parts = StringUtils.csvLineToList(StringUtils.getInnerString(nodeContent, "(", ")"));
			this.entityTypeExpression = parts.get(0);
			for (int i = parts.size() - 1; i >= 0; i--)
			{
				if (parts.get(i).trim().startsWith("menu="))
				{
					menuCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("view="))
				{
					viewCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("title="))
				{
					titleExpression = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("entity="))
				{
					entityTypeExpression = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("newwindow="))
				{
					targetExpression = StringUtils.substringAfter(parts.remove(i), "=");
				}
			}
			if (parts.size() > 1)
				this.titleExpression = parts.get(1);
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder innerBuilder = new StringBuilder();
			renderSubNodes(innerBuilder, helper, currentLineNumber);
			String compressedFields = innerBuilder.toString();
			compressedFields = CommonConstants.BASE64_PREFIX + ObjectToURLUtil.encode(compressedFields);
			String entityType = evalConstantOrVar(entityTypeExpression, helper, currentLineNumber, insideCreator(), this);
			String guiServerURL = calcGUIURL();
			String resolvedMenu = null;
			if (ObjectChecker.isNotEmptyOrNull(menuCode))
				resolvedMenu = evalConstantOrVar(menuCode, helper, currentLineNumber, insideCreator(), this);
			String resolvedViewCode = null;
			if (ObjectChecker.isNotEmptyOrNull(viewCode))
				resolvedViewCode = evalConstantOrVar(viewCode, helper, currentLineNumber, insideCreator(), this);

			String targetWindow = "";

			if (ObjectChecker.isNotEmptyOrNull(targetExpression))
			{
				String target = evalConstantOrVar(targetExpression, helper, currentLineNumber, insideCreator(), this);
				if (ObjectChecker.isNotEmptyOrNull(target))
				{
					if (ObjectChecker.areEqual(target, "popup"))
						targetWindow = target;
					else
						targetWindow = " target='_blank' ";
				}

			}

			String tokens = PlaceTokens
					.tokenize(PREFIX_EDIT, pair(PRM_ENTITY_TYPE, entityType), pair(PRM_FIELDS, compressedFields), pair(PRM_VIEW, ViewMode.NEW),
							pair(PRM_MENU_HASH, resolvedMenu), pair(PRM_LIST_VIEW_NAME, resolvedViewCode));
			String link = guiServerURL + tokens;
			String title = evalConstantOrVar(titleExpression, helper, currentLineNumber, insideCreator(), this);
			if (isDirectLink())
			{
				if (ObjectChecker.isNotEmptyOrNull(targetWindow))
				{
					if (targetWindow.equals("popup"))
					{
						builder.append(CommonConstants.OPEN_LINK_IN_POPUP);
						link = tokens;
					}
					else
					{
						builder.append(CommonConstants.OPEN_LINK_IN_NEW_WINDOW);
					}
				}
				builder.append(link);
			}
			else
			{
				builder.append("<a href='").append(link).append("' ").append(targetWindow).append(">").append(title).append("</a>");
			}
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endcreator");
		}
	}

	interface IReportLinkSubNode
	{
		void render(ReportQuestions questions, RendererPropertyResolver helper, int currentLineNumber);
	}

	private class ReportLinkNode extends RenderNode
	{
		private final String nodeContent;
		private String reportCode;
		private String runType;
		private String plainLink;
		private String title;
		private String targetExpression;
		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(reportCode, runType, title);
		}

		public ReportLinkNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			List<String> parts = StringUtils.csvLineToList(StringUtils.getInnerString(nodeContent, "(", ")"));
			for (int i = parts.size() - 1; i >= 0; i--)
			{
				if (parts.get(i).trim().toLowerCase().startsWith("reportcode="))
					reportCode = StringUtils.substringAfter(parts.remove(i), "=");
				else if (parts.get(i).trim().toLowerCase().startsWith("runtype="))
					runType = StringUtils.substringAfter(parts.remove(i), "=");
				else if (parts.get(i).trim().toLowerCase().startsWith("plainlink="))
					plainLink = StringUtils.substringAfter(parts.remove(i), "=");
				else if (parts.get(i).trim().toLowerCase().startsWith("title="))
					title = StringUtils.substringAfter(parts.remove(i), "=");
				else if (parts.get(i).trim().toLowerCase().startsWith("newwindow="))
					targetExpression = StringUtils.substringAfter(parts.remove(i), "=");
			}
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			String runType = evalConstantOrVar(this.runType, helper, currentLineNumber, false, this);
			Object reportCodeOrId = evalConstantOrVarToObj(this.reportCode, helper, currentLineNumber);
			ReportQuestions questions = ServerNamaRep.repLinkByCode(new HashMap<>(), reportCodeOrId);
			for (RenderNode node : getNodes())
			{
				if (node instanceof IReportLinkSubNode)
				{
					((IReportLinkSubNode) node).render(questions, helper, currentLineNumber);
				}
				else if (node instanceof TextNode)
				{
					//allow text nodes but do not handle them, consider them sort of comment nodes
				}
				else
				{
					throw new UnsupportedOperationException("The node " + node.describe() + " can not be inside " + describe());
				}
			}
			String targetWindow = "";
			if (ObjectChecker.isNotEmptyOrNull(targetExpression))
			{
				if (ObjectChecker.isNotEmptyOrNull(evalConstantOrVar(targetExpression, helper, currentLineNumber, insideCreator(), this)))
					targetWindow = " target='_blank' ";
			}
			Function<String, String> toLink = questions::toResultLink;
			if (ObjectChecker.areEqual(runType, "launch"))
				toLink = questions::toLaunchLink;
			String titleValue = ObjectChecker.toStringOrEmpty(
					TranslationUtil.translateArgument(TranslationUtil.getCurrentLanguage(), questions.getReportReference()));
			if (ObjectChecker.isNotEmptyOrNull(title))
				titleValue = evalConstantOrVar(title, helper, currentLineNumber, false, this);
			if (isPlainLink())
				builder.append(toLink.apply(calcGUIURL()));
			else
				builder.append("<a href='").append(toLink.apply(calcGUIURL())).append("'").append(targetWindow).append(">").append(titleValue).append("</a>");
		}

		private boolean isPlainLink()
		{
			return isTrue(plainLink);
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endreportlink");
		}
	}

	private static boolean isTrue(String s)
	{
		if (s == null || s.isEmpty())
			return false;
		return ObjectChecker.isTrue(s) || ObjectChecker.isTrue(s.replace("\"", "")) || ObjectChecker.isTrue(s.replace("'", ""));
	}

	private class ReportLinkParamMultiValueNode extends RenderNode implements IReportLinkSubNode
	{
		private final String nodeContent;

		public ReportLinkParamMultiValueNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			renderSubNodes(builder, helper, currentLineNumber);
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return "endmutlivalue".equalsIgnoreCase(nodeContent);
		}

		@Override
		public void render(ReportQuestions questions, RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder value = new StringBuilder();
			render(value, helper, currentLineNumber);
			questions.v(value.toString());
		}
	}

	private class ShortenURLNode extends RenderNode
	{
		private final String nodeContent;
		private String server;
		private String signature;

		public ShortenURLNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			List<String> parts = StringUtils.csvLineToList(StringUtils.getInnerString(nodeContent, "(", ")"));
			for (int i = parts.size() - 1; i >= 0; i--)
			{
				if (parts.get(i).trim().startsWith("server="))
					server = StringUtils.substringAfter(parts.remove(i), "=");
				else if (parts.get(i).trim().startsWith("signature="))
					signature = StringUtils.substringAfter(parts.remove(i), "=");
			}
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			String server = evalConstantOrVar(this.server, helper, currentLineNumber, false, this);
			String signature = evalConstantOrVar(this.signature, helper, currentLineNumber, false, this);
			StringBuilder renderedUrl = new StringBuilder();
			renderSubNodes(renderedUrl, helper, currentLineNumber);
			builder.append(shortenURL(server, signature, renderedUrl.toString()));
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endshortenurl");
		}
	}

	public static String shortenURL(String server, String signature, String url)
	{
		String shortURL;
		String requestBody = "action=shorturl&url={0}&format=simple&signature={1}";
		if (!server.endsWith("/"))
			server += "/";

		try
		{
			requestBody = MessageFormat.format(requestBody, URLEncoder.encode(url.toString(), "UTF-8"),
					URLEncoder.encode(signature, "UTF-8"));
			URL url1 = new URL(server + "yourls-api.php");
			HttpURLConnection connection = (HttpURLConnection) url1.openConnection();
			connection.setRequestMethod("POST");
			connection.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
			connection.setRequestProperty("User-Agent", "PostmanRuntime/7.29.0");
			connection.setDoOutput(true);
			try (OutputStream os = connection.getOutputStream())
			{
				os.write(requestBody.getBytes(StandardCharsets.UTF_8));
			}
			int responseCode = connection.getResponseCode();
			if (responseCode == 200 || responseCode == 400)
			{
				//YOURLS returns 400 if the url previously exists, but sends the URL itself in errorStream
				try (InputStream is = responseCode == 200 ? connection.getInputStream() : connection.getErrorStream())
				{
					 shortURL = IOUtils.toString(is);
				}
			}
			else
			{
				throw new RuntimeException("Error: " + responseCode);
			}
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}
		return shortURL;
	}

	private class ReportLinkParamNameNode extends RenderNode implements IReportLinkSubNode
	{
		private final String nodeContent;
		private final String fieldExpression;

		public String getFieldExpression()
		{
			return fieldExpression;
		}

		private ReportLinkParamNameNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.fieldExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			throw new UnsupportedOperationException("The node " + describe() + " is misplaced.");
		}

		@Override
		public void render(ReportQuestions questions, RendererPropertyResolver helper, int currentLineNumber)
		{
			questions.p(evalConstantOrVarToObj(getFieldExpression(), helper, currentLineNumber));
		}
	}

	private class ReportLinkParamValueNode extends RenderNode implements IReportLinkSubNode
	{
		private final String nodeContent;
		private final String fieldExpression;

		public String getFieldExpression()
		{
			return fieldExpression;
		}

		private ReportLinkParamValueNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.fieldExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			throw new UnsupportedOperationException("The node " + describe() + " is misplaced.");
		}

		@Override
		public void render(ReportQuestions questions, RendererPropertyResolver helper, int currentLineNumber)
		{
			if (!fieldExpression.startsWith("\"") || !fieldExpression.endsWith("\""))
			{
				String[] parts = fieldExpression.split(",");
				Object[] args = new Object[parts.length];
				for (int i = 0; i < parts.length; i++)
				{
					args[i] = evalConstantOrVarToObj(parts[i], helper, currentLineNumber);
				}
				questions.v(args);
			}
			else
			{
				questions.v(evalConstantOrVarToObj(fieldExpression, helper, currentLineNumber));
			}
		}
	}

	private class ReportLinkParamRefNode extends RenderNode implements IReportLinkSubNode
	{
		private final String nodeContent;
		private String entityTypeExp;
		private String idExp;
		private String codeExp;
		private String name1Exp;
		private String name2Exp;

		private ReportLinkParamRefNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			String fieldExpression = StringUtils.getInnerString(nodeContent, "(", ")");
			String[] parts = fieldExpression.split(",");
			for (String part : parts)
			{
				if (part.toLowerCase().startsWith("code="))
					codeExp = StringUtils.substringAfter(part, "=");
				else if (part.toLowerCase().startsWith("name1="))
					name1Exp = StringUtils.substringAfter(part, "=");
				else if (part.toLowerCase().startsWith("name2="))
					name2Exp = StringUtils.substringAfter(part, "=");
				else if (part.toLowerCase().startsWith("entitytype="))
					entityTypeExp = StringUtils.substringAfter(part, "=");
				else if (part.toLowerCase().startsWith("id="))
					idExp = StringUtils.substringAfter(part, "=");
			}
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			throw new UnsupportedOperationException("The node " + describe() + " is misplaced.");
		}

		@Override
		public void render(ReportQuestions questions, RendererPropertyResolver helper, int currentLineNumber)
		{
			String entityType = evalConstantOrVar(entityTypeExp, helper, currentLineNumber, false, this);
			Object id = evalConstantOrVarToObj(idExp, helper, currentLineNumber);
			String code = evalConstantOrVar(codeExp, helper, currentLineNumber, false, this);
			EntityReferenceData referenceData = NaMaLayersConnector.getInstance()
					.findByIdOrCode(entityType, ObjectChecker.getFirstNotEmptyObj(id, code));
			if (referenceData == null && ObjectChecker.isNotEmptyOrNull(entityType) && ObjectChecker.isAnyNotEmptyOrNull(code, id))
			{
				String name1 = evalConstantOrVar(name1Exp, helper, currentLineNumber, false, this);
				String name2 = evalConstantOrVar(name2Exp, helper, currentLineNumber, false, this);
				referenceData = new EntityReferenceData(entityType, ObjectChecker.toStringOrEmpty(ReportQuestions.preProcessValue(id)), code, code,
						code,
						name1, name2);
			}
			questions.v(referenceData);
		}
	}

	public class ItemSalesPriceNode extends RenderNode
	{
		private final String nodeContent;
		private String itemIdOrCode;
		private String customerSupplierIdOrCode;
		private String uomCodeOrId;
		private String qty;
		private String classificationIdOrCode;
		private String date;
		private String legalEntityIdOrCode;
		private String sectorIdOrCode;
		private String branchIdOrCode;
		private String analysisSetIdOrCode;
		private String departmentIdOrCode;
		private String revisionIdCode;
		private String colorCode;
		private String sizeCode;
		private String priceClassifier1IdOrCode;
		private String priceClassifier2IdOrCode;
		private String priceClassifier3IdOrCode;
		private String priceClassifier4IdOrCode;
		private String priceClassifier5IdOrCode;
		private String decimalPlaces;
		private String fieldToDisplay;

		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(itemIdOrCode, customerSupplierIdOrCode, uomCodeOrId, qty, classificationIdOrCode, date, legalEntityIdOrCode,
					sectorIdOrCode, branchIdOrCode, analysisSetIdOrCode, departmentIdOrCode, revisionIdCode, colorCode, sizeCode,
					priceClassifier1IdOrCode, priceClassifier2IdOrCode, priceClassifier3IdOrCode, priceClassifier4IdOrCode, priceClassifier5IdOrCode,
					decimalPlaces, fieldToDisplay);
		}

		public String getNodeContent()
		{
			return nodeContent;
		}

		public String getItemIdOrCode()
		{
			return itemIdOrCode;
		}

		public String getCustomerSupplierIdOrCode()
		{
			return customerSupplierIdOrCode;
		}

		public String getUomCodeOrId()
		{
			return uomCodeOrId;
		}

		public String getQty()
		{
			return qty;
		}

		public String getClassificationIdOrCode()
		{
			return classificationIdOrCode;
		}

		public String getDate()
		{
			return date;
		}

		public String getLegalEntityIdOrCode()
		{
			return legalEntityIdOrCode;
		}

		public String getSectorIdOrCode()
		{
			return sectorIdOrCode;
		}

		public String getBranchIdOrCode()
		{
			return branchIdOrCode;
		}

		public String getAnalysisSetIdOrCode()
		{
			return analysisSetIdOrCode;
		}

		public String getDepartmentIdOrCode()
		{
			return departmentIdOrCode;
		}

		public String getRevisionIdCode()
		{
			return revisionIdCode;
		}

		public String getColorCode()
		{
			return colorCode;
		}

		public String getSizeCode()
		{
			return sizeCode;
		}

		public String getPriceClassifier1IdOrCode()
		{
			return priceClassifier1IdOrCode;
		}

		public String getPriceClassifier2IdOrCode()
		{
			return priceClassifier2IdOrCode;
		}

		public String getPriceClassifier3IdOrCode()
		{
			return priceClassifier3IdOrCode;
		}

		public String getPriceClassifier4IdOrCode()
		{
			return priceClassifier4IdOrCode;
		}

		public String getPriceClassifier5IdOrCode()
		{
			return priceClassifier5IdOrCode;
		}

		public String getDecimalPlaces()
		{
			return decimalPlaces;
		}

		public String getFieldToDisplay()
		{
			return fieldToDisplay;
		}

		private ItemSalesPriceNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			List<String> parts = StringUtils.csvLineToList(StringUtils.getInnerString(nodeContent, "(", ")"));
			for (int i = parts.size() - 1; i >= 0; i--)
			{
				if (isCodeOrId(parts.get(i), "item"))
				{
					itemIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "customer") || isCodeOrId(parts.get(i), "supplier"))
				{
					customerSupplierIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "uom"))
				{
					uomCodeOrId = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("qty="))
				{
					qty = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("decimalPlaces="))
				{
					decimalPlaces = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "classification"))
				{
					classificationIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("date="))
				{
					date = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "legalEntity"))
				{
					legalEntityIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "sector"))
				{
					sectorIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "branch"))
				{
					branchIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "analysisSet"))
				{
					analysisSetIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "department"))
				{
					departmentIdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "revision"))
				{
					revisionIdCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("colorCode="))
				{
					colorCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("sizeCode="))
				{
					sizeCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "priceClassifier1"))
				{
					priceClassifier1IdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "priceClassifier2"))
				{
					priceClassifier2IdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "priceClassifier3"))
				{
					priceClassifier3IdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "priceClassifier4"))
				{
					priceClassifier4IdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (isCodeOrId(parts.get(i), "priceClassifier5"))
				{
					priceClassifier5IdOrCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("fieldToDisplay="))
				{
					fieldToDisplay = StringUtils.substringAfter(parts.remove(i), "=");
				}
			}
		}

		private boolean isCodeOrId(String itemCodeOrId, String item)
		{
			return itemCodeOrId.trim().startsWith(item + "IdOrCode=") || itemCodeOrId.trim().startsWith(item + "CodeOrId=");
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(helper.calcSalesPriceForCustomer(this, currentLineNumber));
		}

	}

	public class ItemPurchasePriceNode extends ItemSalesPriceNode
	{
		private ItemPurchasePriceNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(helper.calcPurchasePriceForSupplier(this, currentLineNumber));
		}

	}

	private String calcGUIURL()
	{
		if (ObjectChecker.isNotEmptyOrNull(appURL))
			return appURL;
		String guiServerURL = NaMaLayersConnector.getInstance().getGuiServerURL();
		if (isDirectLink() || isShortLinks())
			guiServerURL = "";
		return guiServerURL;
	}

	private String evalConstantOrVar(String expression, RendererPropertyResolver helper, int currentLineNumber, boolean insideCreator,
			RenderNode calledFrom)
	{
		if (ObjectChecker.isEmptyOrNull(expression))
			return "";
		try
		{
			new BigDecimal(expression);
			return expression;
		}
		catch (Exception ignored)
		{

		}
		if (expression.startsWith("\""))
			return StringUtils.getInnerString(expression, "\"", "\"");
		if (expression.startsWith("@@") && counters.containsKey(expression.substring("@@".length())))
			return ObjectChecker.toStringOrEmpty(counters.get(expression.substring("@@".length())));
		return helper.resolve(expression, currentLineNumber, insideCreator, calledFrom);
	}

	private Object evalConstantOrVarToObj(String expression, RendererPropertyResolver helper, int currentLineNumber)
	{
		if (ObjectChecker.isEmptyOrNull(expression))
			return null;
		try
		{
			return new BigDecimal(expression);
		}
		catch (Exception ignored)
		{

		}
		if (expression.startsWith("\""))
			return StringUtils.getInnerString(expression, "\"", "\"");
		if (expression.startsWith("@@") && counters.containsKey(expression.substring("@@".length())))
			return counters.get(expression.substring("@@".length()));
		return helper.resolveObject(expression, currentLineNumber);
	}

	public static ComplexRenderer parse(String text)
	{
		return new ComplexRenderer(text).parse();
	}

	public ComplexRenderer parse()
	{
		currentNode = rootNode = new RootNode();
		text = ObjectChecker.toStringOrEmpty(text);
		if (text.contains("<notempo/>"))
		{
			TextNode node = new TextNode();
			currentNode.addSubNode(node);
			node.setTextPart(text.replace("<notempo/>", ""));
			return this;
		}
		if (text.contains("<useCSSFriendlyBrackets/>"))
		{
			this.useCSSFriendlyBrackets = true;
			text = text.replace("<useCSSFriendlyBrackets/>", "");
		}
		for (int i = 0; i < text.length(); i++)
		{
			if (text.charAt(i) == '\\')
				i++;
			if (isOpenBracket(i))
			{
				if (useCSSFriendlyBrackets)
					i++;
				int closeBracketIndex = getCloseBracketIndex(i);
				if (closeBracketIndex < 0)
					throw new RuntimeException("Could not find closing bracket in the following text: " + text.substring(i));
				String nodeContent = text.substring(i + 1, closeBracketIndex);
				if (nodeContent.toLowerCase().startsWith("#"))
					nodeContent = nodeContent.substring(1);
				i = closeBracketIndex;
				if (useCSSFriendlyBrackets)
					i++;
				try
				{
					handleNode(nodeContent);
				}
				catch (Exception e)
				{
					Object r1 = HELPER.createFailure("Error while parsing template {0}", text);
					Object r2 = HELPER.createFailure("The part {0} is invalid, please see previous error for the full template",
							"{" + nodeContent + "}");
					throw HELPER.exception(e, r1, r2);
				}
			}
			else
			{
				if (!currentNode.canHaveText())
				{
					TextNode textNode = new TextNode();
					currentNode.addSubNode(textNode);
				}
				currentNode.consume(text.charAt(i));
			}
		}
		return this;
	}

	private int getCloseBracketIndex(int i)
	{
		String closeBracket = "}";
		if (useCSSFriendlyBrackets)
			closeBracket = "}%";
		int closeBracketIndex = text.indexOf(closeBracket, i);
		if (!useCSSFriendlyBrackets && closeBracketIndex > 0)
		{
			if (text.charAt(closeBracketIndex - 1) == '\\')
				return getCloseBracketIndex(closeBracketIndex + 1);
		}
		return closeBracketIndex;
	}

	private boolean isOpenBracket(int i)
	{
		if (useCSSFriendlyBrackets)
		{
			if (i < text.length() - 1 && text.substring(i, i + 2).equals("%{"))
				return true;
			else
				return false;
		}
		if (text.charAt(i) == '{')
		{
			if (i > 0)
			{
				if (text.charAt(i - 1) == '\\')
					return false;
				return true;
			}
			return true;
		}
		return false;
	}

	private void handleNode(String nodeContent)
	{
		if (nodeContent.toLowerCase().startsWith("end"))
		{
			RenderNode node = currentNode;
			while (node != null)
			{
				if (node.isMyEnder(nodeContent))
				{
					node.end(nodeContent);
					return;
				}
				node = node.getParentNode();
			}
		}
		assert currentNode != null;
		if (nodeContent.toLowerCase().startsWith("loop("))
		{
			currentNode.addSubNode(new LoopNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("else") && handleIfNode(nodeContent, true))
		{
			noOp();
		}
		else if (handleIfNode(nodeContent, false))
		{
			noOp();
		}
		else if (nodeContent.toLowerCase().startsWith("incrementcounter("))
		{
			currentNode.addSubNode(new IncrementCounterNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("decrementcounter("))
		{
			currentNode.addSubNode(new DecrementCounterNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("countervalue("))
		{
			currentNode.addSubNode(new CounterValueNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("padleft("))
		{
			currentNode.addSubNode(new PadLeftNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("padright("))
		{
			currentNode.addSubNode(new PadRightNode(nodeContent));
		}
		else if (nodeContent.equals("tempo"))
		{
			currentNode.addSubNode(new TempoNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("appurl("))
		{
			currentNode.addSubNode(new AppUrlNode(nodeContent));
		}
		else if (nodeContent.equals("enter"))
		{
			currentNode.addSubNode(new EnterNode());
		}
		else if (nodeContent.equals("directlinks"))
		{
			currentNode.addSubNode(new DirectLinksNode());
		}
		else if (nodeContent.equals("openinnewwindow") || nodeContent.equals("newwindow"))
		{
			currentNode.addSubNode(new OpenInNewWindowNode());
		}
		else if (nodeContent.equals("shortlinks"))
		{
			currentNode.addSubNode(new ShortLinksNode());
		}
		else if (nodeContent.equals("dynamiccriteria"))
		{
			currentNode.addSubNode(new DynamicCriteriaNode());
		}
		else if (nodeContent.equals("comment"))
		{
			currentNode.addSubNode(new CommentNode());
		}
		else if (nodeContent.toLowerCase().startsWith("substring("))
		{
			currentNode.addSubNode(new SubstringNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("left("))
		{
			currentNode.addSubNode(new LeftNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("right("))
		{
			currentNode.addSubNode(new RightNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("header("))
		{
			currentNode.addSubNode(new GroupHeaderNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("footer("))
		{
			currentNode.addSubNode(new GroupFooterNode(nodeContent));
		}
		else if (nodeContent.equals("@rownumber"))
		{
			currentNode.addSubNode(new RowNumberNode());
		}

		else if (nodeContent.toLowerCase().startsWith("creator("))
		{
			currentNode.addSubNode(new CreatorNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("f("))
		{
			currentNode.addSubNode(new CreatorFieldNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("v("))
		{
			currentNode.addSubNode(new CreatorValueNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("callguiaction("))
		{
			currentNode.addSubNode(new CreatorCallGUIActionNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("creatorvalue"))
		{
			currentNode.addSubNode(new CreatorMultiValueNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("r("))
		{
			currentNode.addSubNode(new CreatorRowNode(nodeContent));
		}
		else if (nodeContent.equals("table"))
		{
			currentNode.addSubNode(new TableNode(nodeContent));
		}
		else if (nodeContent.equals("msg"))
		{
			currentNode.addSubNode(new MsgNode(nodeContent));
		}
		else if (nodeContent.equals("openmsg"))
		{
			currentNode.addSubNode(new StartMsgNode());
		}
		else if (nodeContent.equals("closemsg"))
		{
			currentNode.addSubNode(new EndMsgNode());
		}
		else if (nodeContent.equals("opentable"))
		{
			currentNode.addSubNode(new OpenTableNode());
		}
		else if (nodeContent.equals("closetable"))
		{
			currentNode.addSubNode(new CloseTableNode());
		}
		else if (nodeContent.equals("sendto"))
		{
			currentNode.addSubNode(new SendToNode(nodeContent));
		}
		else if (nodeContent.equals("subject"))
		{
			currentNode.addSubNode(new SubjectNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("emailattachment("))
		{
			currentNode.addSubNode(new EmailAttachmentNode(nodeContent));
		}
		else if (nodeContent.equals("row"))
		{
			currentNode.addSubNode(new TableRowNode(nodeContent));
		}
		else if (nodeContent.equals("cell"))
		{
			currentNode.addSubNode(new TableCellNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("recordimage("))
		{
			currentNode.addSubNode(new RecordImageNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("recordimagelink("))
		{
			currentNode.addSubNode(new RecordImageLinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("attachmentimage("))
		{
			currentNode.addSubNode(new AttachmentImageNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("attachmentlink("))
		{
			currentNode.addSubNode(new AttachmentLinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("titledlink("))
		{
			currentNode.addSubNode(new TitledLinkNode(nodeContent));
		}
		else if (nodeContent.equals("approvelink") || nodeContent.toLowerCase().startsWith("approvelink("))
		{
			currentNode.addSubNode(new ApproveLinkNode(nodeContent));
		}
		else if (nodeContent.equals("rejectlink") || nodeContent.toLowerCase().startsWith("rejectlink("))
		{
			currentNode.addSubNode(new RejectLinkNode(nodeContent));
		}
		else if (nodeContent.equals("returnlink") || nodeContent.toLowerCase().startsWith("returnlink("))
		{
			currentNode.addSubNode(new ReturnLinkNode(nodeContent));
		}
		else if (nodeContent.equals("escalatelink") || nodeContent.toLowerCase().startsWith("escalatelink("))
		{
			currentNode.addSubNode(new EscalateToSupervisorLinkNode(nodeContent));
		}
		else if (nodeContent.equals("returntoprevioussteplink") || nodeContent.toLowerCase().startsWith("returntoprevioussteplink("))
		{
			currentNode.addSubNode(new ReturnToPreviousStepLinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("itemprice("))
		{
			currentNode.addSubNode(new ItemSalesPriceNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("itempurchaseprice("))
		{
			currentNode.addSubNode(new ItemPurchasePriceNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("tafqeet("))
		{
			currentNode.addSubNode(new TafqeetNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("round("))
		{
			currentNode.addSubNode(new RoundNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("formatnumber("))
		{
			currentNode.addSubNode(new FormatNumberNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("formatdate("))
		{
			currentNode.addSubNode(new FormatDateNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("tafqeetar("))
		{
			currentNode.addSubNode(new TafqeetArNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("tafqeeten("))
		{
			currentNode.addSubNode(new TafqeetEnNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("link("))
		{
			currentNode.addSubNode(new LinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("flow("))
		{
			currentNode.addSubNode(new EntityFlowLinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("onlinepayment("))
		{
			currentNode.addSubNode(new OnlinePaymentLinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("time("))
		{
			currentNode.addSubNode(new TimeNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("decimaltotime("))
		{
			currentNode.addSubNode(new DecimalToTimeNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("translate("))
		{
			currentNode.addSubNode(new TranslateNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("translatear("))
		{
			currentNode.addSubNode(new TranslateArNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("translateen("))
		{
			currentNode.addSubNode(new TranslateEnNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("reportlink("))
		{
			currentNode.addSubNode(new ReportLinkNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("paramname("))
		{
			currentNode.addSubNode(new ReportLinkParamNameNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("paramvalue("))
		{
			currentNode.addSubNode(new ReportLinkParamValueNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("paramrefvalue("))
		{
			currentNode.addSubNode(new ReportLinkParamRefNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("parammultivalue"))
		{
			currentNode.addSubNode(new ReportLinkParamMultiValueNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("shortenurl("))
		{
			currentNode.addSubNode(new ShortenURLNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("httprequest"))
		{
			currentNode.addSubNode(new HttpRequestNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("headername"))
		{
			currentNode.addSubNode(new HttpRequestHeaderNameNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("headervalue"))
		{
			currentNode.addSubNode(new HttpRequestHeaderValueNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("bodypartname"))
		{
			currentNode.addSubNode(new HttpRequestBodyPartNameNode(nodeContent));
		}
		else if (nodeContent.toLowerCase().startsWith("bodypart("))
		{
			currentNode.addSubNode(new HttpRequestBodyPartNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("bodypartvalue"))
		{
			currentNode.addSubNode(new HttpRequestBodyPartValueNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("paramname"))
		{
			currentNode.addSubNode(new HttpRequestParameterNameNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("paramvalue"))
		{
			currentNode.addSubNode(new HttpRequestParameterValueNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("fullbody"))
		{
			currentNode.addSubNode(new HttpRequestBodyNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("requestmethod"))
		{
			currentNode.addSubNode(new HttpRequestMethodNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("contenttype"))
		{
			currentNode.addSubNode(new HttpRequestContentTypeNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("charset"))
		{
			currentNode.addSubNode(new HttpRequestCharSetNode(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("requestdescription1"))
		{
			currentNode.addSubNode(new HttpRequestDescription1Node(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("requestdescription2"))
		{
			currentNode.addSubNode(new HttpRequestDescription2Node(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("requestrelatedtoid1"))
		{
			currentNode.addSubNode(new HttpRequestRelatedToId1Node(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("requestrelatedtoid2"))
		{
			currentNode.addSubNode(new HttpRequestRelatedToId2Node(nodeContent));
		}
		else if (nodeContent.equalsIgnoreCase("requesturl"))
		{
			currentNode.addSubNode(new HttpRequestUrlNode(nodeContent));
		}
		else if (ObjectChecker.isAnyEqualToFirst(nodeContent, "end", "endtable", "endloop", "endrow", "endlink", "endcell", "endcreator", "endheader",
				"endfooter", "endsendto", "endsubject", "endmsg", "endcomment", "endif", "endelse", "endvalue", "endpad", "endreportlink",
				"endshortenurl", "endmutlivalue"))
		{
			currentNode.end(nodeContent);
		}
		else
		{
			currentNode.addSubNode(new VarNode(nodeContent));
		}
	}

	private void noOp()
	{

	}

	private boolean handleIfNode(String nodeContent, boolean fromElse)
	{
		String contentWithoutElse = nodeContent;
		if (fromElse)
		{
			contentWithoutElse = nodeContent.substring("else".length()).trim();
			currentNode = findFirstParentIfNode(nodeContent);
			if (contentWithoutElse.isEmpty())//simple else
			{
				currentNode.addSubNode(new ElseNode());
				return true;
			}
		}

		if (contentWithoutElse.toLowerCase().startsWith("iflessoreq(") || contentWithoutElse.toLowerCase().startsWith("if<=(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumberlessoreq(") || contentWithoutElse.toLowerCase().startsWith("ifnumber<=("))
		{
			currentNode.addSubNode(new IfLessOrEqNode(nodeContent));
			return true;
		}
		else if (contentWithoutElse.toLowerCase().startsWith("ifless(") || contentWithoutElse.toLowerCase().startsWith("if<(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumberless(") || contentWithoutElse.startsWith("ifnumber<("))
		{
			currentNode.addSubNode(new IfLessNode(nodeContent));
			return true;
		}
		else if (contentWithoutElse.toLowerCase().startsWith("ifgreateroreq(") || contentWithoutElse.toLowerCase().startsWith("if>=(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumbergreateroreq(") || contentWithoutElse.toLowerCase().startsWith("ifnumber>=("))
		{
			currentNode.addSubNode(new IfGreaterOrEqNode(nodeContent));
			return true;
		}
		else if (contentWithoutElse.toLowerCase().startsWith("ifgreater(") || contentWithoutElse.toLowerCase().startsWith("if>(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumbergreater(") || contentWithoutElse.toLowerCase().startsWith("ifnumber>("))
		{
			currentNode.addSubNode(new IfGreaterNode(nodeContent));
			return true;
		}
		else if (contentWithoutElse.toLowerCase().startsWith("ifequal(") || contentWithoutElse.toLowerCase().startsWith("if=(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumberequal(") || contentWithoutElse.toLowerCase().startsWith("ifnumber=("))
		{
			currentNode.addSubNode(new IfEqualNode(nodeContent));
			return true;
		}
		else if (contentWithoutElse.toLowerCase().startsWith("ifnotequal(") || contentWithoutElse.toLowerCase().startsWith("if!=(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumbernotequal(") || contentWithoutElse.toLowerCase().startsWith("ifnumber!=("))
		{
			currentNode.addSubNode(new IfNotEqualNode(nodeContent));
			return true;
		}
		else if (contentWithoutElse.toLowerCase().startsWith("ifnot(") || contentWithoutElse.toLowerCase().startsWith("if!(")
				|| contentWithoutElse.toLowerCase().startsWith("ifnumbernot(") || contentWithoutElse.toLowerCase().startsWith("ifnumber!("))
		{
			currentNode.addSubNode(new IfNotNode(nodeContent));
			return true;
		}
		else if (nodeContent.toLowerCase().startsWith("if(") || nodeContent.toLowerCase().startsWith("ifnumber("))
		{
			currentNode.addSubNode(new IfNode(nodeContent));
		}
		return false;
	}

	private RenderNode findFirstParentIfNode(String nodeContent)
	{
		RenderNode node = currentNode;
		int loops = 0;
		while (nodeDoesNotAcceptElseNodes(node))
		{
			if (node == null)
				break;
			loops++;
			if (loops > 10000)
				throw new NaMaServiceExcepption("Infinite loop while finding parent if statement for the expression " + nodeContent);
			node = node.parentNode;
		}
		if (node == null)
			throw new NaMaServiceExcepption("Could not find parent if statement for the expression " + nodeContent);
		return node;
	}

	private boolean nodeDoesNotAcceptElseNodes(RenderNode node)
	{
		if (!(node instanceof AbstractIfNode ifNode))
			return true;
		if (!ifNode.acceptsElseNodes())
			return true;
		return ifNode.isElse();
	}

	public StringBuilder render(RendererPropertyResolver helper)
	{
		StringBuilder builder = new StringBuilder();
		rootNode.render(builder, helper, 0);
		if (ObjectChecker.isNotEmptyOrNull(originalAppURL))
			HELPER.setGuiServerURL(originalAppURL);
		return builder;
	}

	@Override
	public String toString()
	{
		StringBuilder builder = new StringBuilder();
		for (RenderNode node : rootNode.getNodes())
		{
			node.toString(builder, 0);
		}
		return builder.toString();
	}

	private class CreatorRowNode extends RenderNode
	{
		private final String nodeContent;
		private final String rowExpression;

		public CreatorRowNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.rowExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(":@").append(evalConstantOrVar(rowExpression, helper, currentLineNumber, insideCreator(), this));
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class TimeNode extends RenderNode
	{
		private final String nodeContent;
		protected final String valueExpression;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(valueExpression);
		}

		public TimeNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.valueExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(ServerNamaRep.timeToString(evalConstantOrVar(valueExpression, helper, currentLineNumber, insideCreator(), this)));
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	public class DecimalToTimeNode extends TimeNode
	{
		public DecimalToTimeNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(ServerNamaRep.decimalToTime(evalConstantOrVar(valueExpression, helper, currentLineNumber, insideCreator(), this)));
		}
	}

	private class TranslateArNode extends TranslateNode
	{

		public TranslateArNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected Language getLang()
		{
			return Language.Arabic;
		}
	}

	private class TranslateEnNode extends TranslateNode
	{

		public TranslateEnNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected Language getLang()
		{
			return Language.English;
		}
	}

	private class TranslateNode extends RenderNode
	{
		private final String nodeContent;
		private final String valueExpression;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(valueExpression);
		}

		public TranslateNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.valueExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(TranslationUtil
					.translate(NaMaText.resource(evalConstantOrVar(valueExpression, helper, currentLineNumber, insideCreator(), this)), getLang()));
		}

		protected Language getLang()
		{
			return HELPER.getLanguage();
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class CreatorFieldNode extends RenderNode
	{
		private final String nodeContent;
		private final String fieldExpression;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(fieldExpression);
		}

		public CreatorFieldNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.fieldExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append("--").append(evalConstantOrVar(fieldExpression, helper, currentLineNumber, insideCreator(), this));
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class CreatorCallGUIActionNode extends RenderNode
	{
		private final String nodeContent;
		private final String systemActionExpression;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(systemActionExpression);
		}

		public CreatorCallGUIActionNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.systemActionExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append("--").append(NestConstants.CALL_GUI_ACTION).append(":=")
					.append(evalConstantOrVar(systemActionExpression, helper, currentLineNumber, insideCreator(), this));
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private abstract class AbsLinkNode extends RenderNode
	{
		private final String nodeContent;
		private final String firstParameter;
		private String secondParameter;
		private String menuCode;
		private String viewCode;
		private String flowCode;
		private String paymentConfigCode;
		private String plainLink;
		private String targetExpression;
		private String imgWidth;
		private String imgHeight;

		public String getFirstParameter()
		{
			return firstParameter;
		}

		public String getPaymentConfigCode()
		{
			return paymentConfigCode;
		}

		public boolean isPlainLink()
		{
			return ObjectChecker.isTrue(plainLink);
		}

		@Override
		public List<String> collectVariables()
		{
			return Arrays.asList(firstParameter, secondParameter, menuCode, viewCode, flowCode, paymentConfigCode, targetExpression);
		}

		public AbsLinkNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			List<String> parts = StringUtils.csvLineToList(StringUtils.getInnerString(nodeContent, "(", ")"));
			firstParameter = parts.get(0);
			for (int i = parts.size() - 1; i > 0; i--)
			{
				if (parts.get(i).trim().startsWith("menu="))
				{
					menuCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().startsWith("view="))
				{
					viewCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("flowcode="))
				{
					flowCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("paymentconfigcode="))
				{
					paymentConfigCode = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("plainlink="))
				{
					plainLink = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("newwindow="))
				{
					targetExpression = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("width="))
				{
					imgWidth = StringUtils.substringAfter(parts.remove(i), "=");
				}
				else if (parts.get(i).trim().toLowerCase().startsWith("height="))
				{
					imgHeight = StringUtils.substringAfter(parts.remove(i), "=");
				}
			}
			if (parts.size() > 1 && !parts.get(1).contains("="))
				secondParameter = parts.get(1);
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			Object first;
			if (firstParameter.startsWith("\""))
				first = evalConstantOrVar(firstParameter, helper, currentLineNumber, insideCreator(), this);
			else
				first = helper.resolveObject(firstParameter, currentLineNumber);
			Object entityType = null;
			Object id = "";
			Object title = "";
			ExtractLinkInfo extractLinkInfo = helper.extractLinkInfo(first, entityType, id, title);
			entityType = extractLinkInfo.getEntityType();
			id = extractLinkInfo.getId();
			title = extractLinkInfo.getTitle();
			if (ObjectChecker.isNotEmptyOrNull(secondParameter))
				id = evalConstantOrVar(secondParameter, helper, currentLineNumber, insideCreator(), this);
			StringBuilder innerBuilder = new StringBuilder();
			renderSubNodes(innerBuilder, helper, currentLineNumber);
			if (ObjectChecker.isNotEmptyOrNull(innerBuilder.toString()))
			{
				title = innerBuilder.toString();
			}
			String resolvedMenuCode = null;
			String resolvedView = null;
			String resolvedFlowCode = null;
			if (ObjectChecker.isNotEmptyOrNull(menuCode))
			{
				resolvedMenuCode = evalConstantOrVar(menuCode, helper, currentLineNumber, insideCreator(), this);
			}
			if (ObjectChecker.isNotEmptyOrNull(viewCode))
			{
				resolvedView = evalConstantOrVar(viewCode, helper, currentLineNumber, insideCreator(), this);
			}
			if (ObjectChecker.isNotEmptyOrNull(flowCode))
			{
				resolvedFlowCode = evalConstantOrVar(flowCode, helper, currentLineNumber, insideCreator(), this);
			}
			String url = calcGUIURL();
			String targetWindow = "";

			if (ObjectChecker.isNotEmptyOrNull(targetExpression))
			{
				if (ObjectChecker.isNotEmptyOrNull(evalConstantOrVar(targetExpression, helper, currentLineNumber, insideCreator(), this)))
					targetWindow = " target='_blank' ";
			}
			LinkCreationParameters creationParameters = new LinkCreationParameters(builder, url, entityType, id, resolvedMenuCode, resolvedView,
					resolvedFlowCode, targetWindow, title);
			creationParameters.setImgHeight(evalConstantOrVar(imgHeight, helper, currentLineNumber, insideCreator(), this));
			creationParameters.setImgWidth(evalConstantOrVar(imgWidth, helper, currentLineNumber, insideCreator(), this));
			renderLink(creationParameters);
		}

		protected void renderLink(LinkCreationParameters linkCreationParameters)
		{
			if (isPlainLink())
			{
				linkCreationParameters.getBuilder().append(linkCreationParameters.getUrl())
						.append(getPlaceToken(linkCreationParameters.getEntityType(), linkCreationParameters.getId(),
								linkCreationParameters.getResolvedMenuCode(), linkCreationParameters.getResolvedView(),
								linkCreationParameters.getResolvedFlowCode()));
				return;
			}
			linkCreationParameters.getBuilder().append("<a href='").append(linkCreationParameters.getUrl())
					.append(getPlaceToken(linkCreationParameters.getEntityType(), linkCreationParameters.getId(),
							linkCreationParameters.getResolvedMenuCode(), linkCreationParameters.getResolvedView(),
							linkCreationParameters.getResolvedFlowCode())).append("' ").append(linkCreationParameters.getTargetWindow()).append(">")
					.append(linkCreationParameters.getTitle()).append("</a>");
		}

		protected String getPlaceToken(Object entityType, Object id, String resolvedMenuCode, String resolvedView, String resolvedFlowCode)
		{
			return PlaceTokens
					.getEditPlaceTokenWithMenu(ObjectChecker.toStringOrEmpty(entityType), ObjectChecker.toStringOrEmpty(id), resolvedMenuCode,
							resolvedView);
		}

	}

	private class RecordImageNode extends AbsLinkNode
	{

		public RecordImageNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected void renderLink(LinkCreationParameters p)
		{
			if (ObjectChecker.isEmptyOrNull(p.getId()))
				return;
			p.getBuilder().append("<img src='").append(PlaceTokens.downloadURLByRecordId(ObjectChecker.toStringOrEmpty(p.getId()),
					ObjectChecker.toStringOrEmpty(p.getEntityType()), DownloadPurpose.PREVIEW, false)).append("' ");
			if (ObjectChecker.isNotEmptyOrNull(p.getImgWidth()))
				p.getBuilder().append(" width='").append(p.getImgWidth()).append("' ");
			if (ObjectChecker.isNotEmptyOrNull(p.getImgHeight()))
				p.getBuilder().append(" height='").append(p.getImgHeight()).append("' ");
			if(ObjectChecker.areAllEmptyOrNull(p.getImgHeight(),p.getImgWidth()))
				p.getBuilder().append(" width='150' ");
			p.getBuilder().append("/>");
		}
	}

	private class AttachmentImageNode extends AbsLinkNode
	{

		public AttachmentImageNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected void renderLink(LinkCreationParameters p)
		{
			if (ObjectChecker.isEmptyOrNull(p.getId()))
				return;
			p.getBuilder().append("<img src='")
					.append(PlaceTokens.downloadURL(ObjectChecker.toStringOrEmpty(p.getId()), ObjectChecker.toStringOrEmpty(p.getEntityType()),
							DownloadPurpose.PREVIEW, false)).append("' ");
			if (ObjectChecker.isNotEmptyOrNull(p.getImgWidth()))
				p.getBuilder().append(" width='").append(p.getImgWidth()).append("' ");
			if (ObjectChecker.isNotEmptyOrNull(p.getImgHeight()))
				p.getBuilder().append(" height='").append(p.getImgHeight()).append("' ");
			p.getBuilder().append("/>");
		}
	}

	private class RecordImageLinkNode extends AbsLinkNode
	{

		public RecordImageLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected void renderLink(LinkCreationParameters p)
		{
			if (ObjectChecker.isEmptyOrNull(p.getId()))
				return;
			p.getBuilder().append(PlaceTokens.downloadURLByRecordId(ObjectChecker.toStringOrEmpty(p.getId()),
					ObjectChecker.toStringOrEmpty(p.getEntityType()), DownloadPurpose.PREVIEW, false));
		}
	}

	private class AttachmentLinkNode extends AbsLinkNode
	{

		public AttachmentLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		protected void renderLink(LinkCreationParameters p)
		{
			if (ObjectChecker.isEmptyOrNull(p.getId()))
				return;
			p.getBuilder().append(PlaceTokens.downloadURL(ObjectChecker.toStringOrEmpty(p.getId()), ObjectChecker.toStringOrEmpty(p.getEntityType()),
					DownloadPurpose.PREVIEW, false));
		}
	}
	public static class ExtractLinkInfo
	{
		private Object first;
		private final Object entityType;
		private final Object id;
		private final Object title;

		public ExtractLinkInfo(Object first, Object entityType, Object id, Object title)
		{
			this.first = first;
			this.entityType = entityType;
			this.id = id;
			this.title = title;
		}

		public Object getEntityType()
		{
			return entityType;
		}

		public Object getId()
		{
			return id;
		}

		public Object getTitle()
		{
			return title;
		}

	}

	private class TitledLinkNode extends AbsLinkNode
	{

		public TitledLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endlink");
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}
	}

	private abstract class AbsApprovalLinkNode extends RenderNode
	{

		private final String nodeContent;

		public AbsApprovalLinkNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			INotificationInfo notificationInfo = helper.calcNotificationInfo(currentLineNumber);
			if (notificationInfo == null)
				throw HELPER.exception(null, HELPER.createFailure("You can only use approval links inside approval templates"));
			builder.append("<a target='_blank' href='");
			builder.append(ServerApprovalUtils
					.approvalUrl(getApprovalGUIURL(), decision(), notificationInfo.calcCandidateEmployeeId(), notificationInfo.calcApprovalLines(),
							notificationInfo.calcApprovedRecordId(), notificationInfo.
									calcApprovedRecordType(), notificationInfo.calcApprovedRecordCode(), notificationInfo.calcApprovalSecret(),
							notificationInfo.calcApprovalStepSeq(), reason()));
			builder.append("'>");
			builder.append(TranslationUtil.translateArgument(TranslationUtil.getCurrentLanguage(), decision()));
			builder.append("</a>");
		}

		private String reason()
		{
			List<String> parts = StringUtils.csvLineToList(StringUtils.getInnerString(nodeContent, "(", ")"));
			for (String part : parts)
			{
				if (StringUtils.substringBefore(part, "=").trim().equals("reason"))
					return StringUtils.substringAfter(part, "=").trim();
			}
			return null;
		}

		protected abstract ApprovalDecision decision();

		protected String getApprovalGUIURL()
		{
			return calcGUIURL();
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class ApproveLinkNode extends AbsApprovalLinkNode
	{

		public ApproveLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected ApprovalDecision decision()
		{
			return ApprovalDecision.Approve;
		}
	}

	private class RejectLinkNode extends AbsApprovalLinkNode
	{

		public RejectLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected ApprovalDecision decision()
		{
			return ApprovalDecision.Reject;
		}
	}

	private class ReturnLinkNode extends AbsApprovalLinkNode
	{

		public ReturnLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected ApprovalDecision decision()
		{
			return ApprovalDecision.Return;
		}
	}

	private class EscalateToSupervisorLinkNode extends AbsApprovalLinkNode
	{

		public EscalateToSupervisorLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected ApprovalDecision decision()
		{
			return ApprovalDecision.EscalateToSupervisor;
		}
	}

	private class ReturnToPreviousStepLinkNode extends AbsApprovalLinkNode
	{

		public ReturnToPreviousStepLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected ApprovalDecision decision()
		{
			return ApprovalDecision.ReturnToPreviousStep;
		}
	}

	private class LinkNode extends AbsLinkNode
	{
		public LinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class EntityFlowLinkNode extends AbsLinkNode
	{
		public EntityFlowLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected String getPlaceToken(Object entityType, Object id, String resolvedMenuCode, String resolvedView, String resolvedFlowCode)
		{
			return PlaceTokens.tokenize("", pair(PRM_FLOW_ID, resolvedFlowCode), pair(PRM_FLOW_TARGET_TYPE, entityType),
					pair(PRM_FLOW_TARGET_ID, id));
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class OnlinePaymentLinkNode extends AbsLinkNode
	{
		public OnlinePaymentLinkNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			Object document = evalConstantOrVarToObj(getFirstParameter(), helper, currentLineNumber);
			Object config = evalConstantOrVarToObj(getPaymentConfigCode(), helper, currentLineNumber);
			String url = helper.createPaymentUrl(document, config);
			if (isPlainLink())
			{
				builder.append(url);
				return;
			}
			builder.append("<a target='_blank' href='");
			builder.append(url);
			builder.append("'>Payment URL</a>");
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class CreatorValueNode extends RenderNode
	{
		private final String nodeContent;
		private final String valueExpression;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(valueExpression);
		}

		public CreatorValueNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.valueExpression = StringUtils.getInnerString(nodeContent, "(", ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			builder.append(":=").append(evalConstantOrVar(valueExpression, helper, currentLineNumber, insideCreator(), this));
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}
	}

	private class CreatorMultiValueNode extends RenderNode
	{
		private final String nodeContent;

		public CreatorMultiValueNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder innerBuilder = new StringBuilder();
			renderSubNodes(innerBuilder, helper, currentLineNumber);
			builder.append(":=").append(innerBuilder);
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return "endvalue".equalsIgnoreCase(nodeContent);
		}
	}

	private abstract class PrefixSuffixCompositeNode extends RenderNode
	{
		protected String nodeContent;

		public PrefixSuffixCompositeNode(String nodeContent)
		{
			super();
			this.nodeContent = nodeContent;
		}

		@Override
		public boolean isLeafNode()
		{
			return false;
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			StringBuilder innerBuilder = new StringBuilder();
			innerBuilder.append(prefix(helper, currentLineNumber));
			renderSubNodes(innerBuilder, helper, currentLineNumber);
			innerBuilder.append(suffix(helper, currentLineNumber));
			builder.append(innerBuilder);
		}

		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "";
		}

		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "";
		}
	}

	private class TableNode extends PrefixSuffixCompositeNode
	{
		public TableNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return OPEN_TABLE;
		}

		@Override
		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "\n</table>";
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endtable");
		}
	}

	private class TableRowNode extends PrefixSuffixCompositeNode
	{
		public TableRowNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean shouldSubNodeEndThis(RenderNode subNode)
		{
			return subNode instanceof TableRowNode;
		}

		@Override
		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "\n<tr>";
		}

		@Override
		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "\n</tr>";
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endrow");
		}
	}

	private class TableCellNode extends PrefixSuffixCompositeNode
	{
		public TableCellNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean shouldSubNodeEndThis(RenderNode subNode)
		{
			return subNode instanceof TableCellNode || subNode instanceof TableRowNode;
		}

		@Override
		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "\n<td>";
		}

		@Override
		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return "\n</td>";
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endcell");
		}
	}

	private class MsgNode extends PrefixSuffixCompositeNode
	{
		public MsgNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return START_NEW_MSG;
		}

		@Override
		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return END_MSG;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endmsg");
		}
	}

	private class SendToNode extends PrefixSuffixCompositeNode
	{
		public SendToNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean shouldSubNodeEndThis(RenderNode subNode)
		{
			return subNode instanceof MsgNode;
		}

		@Override
		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return START_SEND_TO;
		}

		@Override
		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return END_SEND_TO;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endsendto");
		}
	}

	private class SubjectNode extends PrefixSuffixCompositeNode
	{
		public SubjectNode(String nodeContent)
		{
			super(nodeContent);
		}

		@Override
		public boolean shouldSubNodeEndThis(RenderNode subNode)
		{
			return subNode instanceof MsgNode;
		}

		@Override
		protected Object prefix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return START_MSG_SUBJECT;
		}

		@Override
		protected Object suffix(RendererPropertyResolver helper, int currentLineNumber)
		{
			return END_MSG_SUBJECT;
		}

		@Override
		protected boolean isMyEnder(String nodeContent)
		{
			return ObjectChecker.areEqual(nodeContent, "endsubject");
		}
	}

	public class EmailAttachmentNode extends RenderNode
	{
		private final String expression;
		private final String nodeContent;

		@Override
		public List<String> collectVariables()
		{
			return Collections.singletonList(expression);
		}

		public EmailAttachmentNode(String nodeContent)
		{
			this.nodeContent = nodeContent;
			this.expression = StringUtils.substringBeforeLast(StringUtils.substringAfter(nodeContent, "("), ")");
		}

		@Override
		protected String describe()
		{
			return "{" + nodeContent + "}";
		}

		@Override
		public boolean isLeafNode()
		{
			return true;
		}

		@Override
		public void render(StringBuilder builder, RendererPropertyResolver helper, int currentLineNumber)
		{
			String expressionValue = helper.objectToUUIDIfPossible(evalConstantOrVarToObj(expression, helper, currentLineNumber));
			if (ObjectChecker.isEmptyOrNull(expressionValue))
				return;
			builder.append(START_ATTACH_DATA_TAG).append(expressionValue).append(END_ATTACH_DATA_TAG);
		}
	}
}